#!/usr/bin/python3
# coding: utf8

# The Incredible Machine (TIM) image resource reader, version 0.1
# By knt47, licensed under the Creative Commons Zero license

from __future__ import division, print_function, unicode_literals

import itertools, sys

if sys.version_info[0] != 3:
    raise NotImplementedError('Only Python 3 is supported.')
else:
    izip = zip
    xrange = range

from timgres.base.misc import s_opt_transpcol16, s_opt_transpcol256, s_opt_zoom, \
    s_opt_flip, s_opt_pal_file_def, s_opt_pal_file_def_2, s_opt_font_color, \
    s_opt_csv_format, s_opt_preserve_pos_size_singleimg, \
 \
    s_obj_resreadmap, s_obj_resreaddir, s_obj_pngexp2, s_obj_pngexp16, \
    s_obj_pngexp256, s_obj_chunkuncompr, s_obj_pngmerge16, s_obj_pngmerge256, \
    s_obj_pngshapedraw16, s_res_file_special, s_open_res_cback_pal, \
 \
    ext_subf_anm, ext_subf_bmp, ext_subf_fnt, ext_subf_pal, \
    ext_subf_scr, opt_pal_file_def, opt_pal_file_2_def, opt_font_color_def, \
    opt_csv_format_def


from timgres.base.basecl import nt_obj_manager_option, s_opt_type_int, \
    s_opt_type_float, s_opt_type_int_none, s_opt_type_str, \
    s_opt_type_color_none

from timgres.base.rdatamgr import ResDataManager

# * "other" object classes *

from timgres.base.resfile import ResFileReader_Map_T, ResFileReader_Dir

try:
    # TODO, not implemented
    from timgres.extra.dcchunk import ChunkUncompressor as ChunkUncompressor_CliCfg
except ImportError:
    # Module not available, use the base class as fallback
    from timgres.base.dcunbase import ChunkUncompressorBase as ChunkUncompressor_CliCfg

from timgres.base.pngexp import SimplePngExporter, PNG_COL_TYPE_INDEXED
from timgres.base.pngmerge import _BPP_4, _BPP_8, PngMerger_4BPP, PngMerger_8BPP
from timgres.base.pngdraw import PngShapeDraw_4BPP

# * decoder classes *
from timgres.decoder.d_bmpbin import SubFileDecoder_BmpBin
from timgres.decoder.d_fnt import SubFileDecoder_Fnt_1BPP
from timgres.decoder.d_scr import SubFileDecoder_Scr
from timgres.decoder.d_t12pal import SubFileDecoder_Pal_T12
from timgres.decoder.d_t3pal import SubFileDecoder_Pal_T3
from timgres.decoder.d_t12bmp import SubFileDecoder_Bmp_T12
from timgres.decoder.d_t3bmp import SubFileDecoder_Bmp_T3
from timgres.decoder.d_t2anm import SubFileDecoder_Anm_T2

# * ViewerDataObj classes *
from timgres.viewdata.vd_bmp import ViewerDataObj_Bmp, \
    ViewerDataObj_Fnt, ViewerDataObj_Scr
from timgres.viewdata.vd_pal import ViewerDataObj_Pal, open_res_callback_pal
from timgres.viewdata.vd_anm import ViewerDataObj_AnmFrames
from timgres.viewdata.vd_sdata import ViewerDataObj_StoredData

# * TIM/TEMIM "part info" *
try:
    from timgres.extra.t1pinfo import info_checksums, info_frame_sub_image_pos
except ImportError:
    info_checksums = None
    info_frame_sub_image_pos = None

# * exporter classes *

# TODO TODO: add further
from timgres.base.defexp import SingleBinFileExporter, AllBinFilesExporter, \
    BinFileListExporter, BinFilePatternExporter, ImageSingleSubItemExporter, \
    ImageSubItemsExporter, ImageListSubItemsExporter, \
    ImageFilePatternSubItemsExporter, PlainNameListPatternExporter, \
    PlainSubItemNameListExporter, FileInfoCsvExporter, ItemInfoCsvExporter

# --- ResDataManager subclass, "configured" for CLI ---

'''
This module contains the ResDataManager_CliConf class, a sub-class of
    ResDataManager. This class is responsible for the creation of all
    objects, exporters and viewers required for both the CLI and the GUI.
'''

_BPP_1 = 1  # 1 bits per pixel, 2 colors
# now imported from pngmerge
# _BPP_4 = 4  # 4 bits per pixel, max. 16 colors
# _BPP_8 = 8  # 8 bits per pixel, max. 256 colors

file_ext_descr_anm = 'Animation files (.ANM)'
file_ext_descr_bmp = 'Bitmap files (.BMP)'
file_ext_descr_fnt = 'Font files (.FNT)'
file_ext_descr_scr = 'Full-screen files (.SCR)'
file_ext_descr_pal = 'Palette files (.PAL)'


class ResDataManager_CliConf(ResDataManager):
    '''
    A ResDataManager subclass, used by the CLI part of the program.
        Creates all Resoucre Manager objects that are used in the CLI
        part of the program, and much of the objects used in the GUI
        version.
    '''

    # NOTE: options_default is empty in parent !
    # options_default = TreeObjectsManager.options_default + \

    options_default = (
        nt_obj_manager_option(s_opt_transpcol16, s_opt_type_int_none, None),
        nt_obj_manager_option(s_opt_transpcol256, s_opt_type_int_none, None),
        nt_obj_manager_option(s_opt_zoom, s_opt_type_int, 1),
        nt_obj_manager_option(s_opt_flip, s_opt_type_int, 0),
        nt_obj_manager_option(s_opt_pal_file_def, s_opt_type_str,
                              opt_pal_file_def),
        nt_obj_manager_option(s_opt_pal_file_def_2, s_opt_type_str,
                              opt_pal_file_2_def),
        # FOXME, font color should not be None, default is black
        nt_obj_manager_option(s_opt_font_color, s_opt_type_color_none,
                              opt_font_color_def),
        nt_obj_manager_option(s_opt_csv_format, s_opt_type_str,
                              opt_csv_format_def),
        nt_obj_manager_option(s_opt_preserve_pos_size_singleimg,
                              s_opt_type_int, False),
    )

    def create_default_objects_resdata(self):
        '''
        Calls the create_default_objects_resdata method of the parent class.
        Creates all objects that are both accessible in the CLI and the GUI version.
        '''

        # NOTE: create_default_objects_treedata does nothing in parent class
        ResDataManager.create_default_objects_resdata(self)

        # * "other" objects *
        self.add_object(s_obj_resreadmap, ResFileReader_Map_T())
        self.add_object(s_obj_resreaddir, ResFileReader_Dir())

        obj_pngexp2 = SimplePngExporter()
        obj_pngexp2.set_bit_depth_color_type(_BPP_1, PNG_COL_TYPE_INDEXED)
        self.add_object(s_obj_pngexp2, obj_pngexp2)

        obj_pngexp16 = SimplePngExporter()
        obj_pngexp16.set_bit_depth_color_type(_BPP_4, PNG_COL_TYPE_INDEXED)
        self.add_object(s_obj_pngexp16, obj_pngexp16)

        obj_pngexp256 = SimplePngExporter()
        obj_pngexp256.set_bit_depth_color_type(_BPP_8, PNG_COL_TYPE_INDEXED)
        self.add_object(s_obj_pngexp256, obj_pngexp256)

        self.add_object(s_obj_chunkuncompr, ChunkUncompressor_CliCfg())

        # Note: PNG merger objects are stored in the "object" dict, but
        #     require 'parent' like decoders and exporters.
        #     set_parent() must be called after object creation

        png_merger_4bpp = PngMerger_4BPP()
        png_merger_4bpp.set_parent(self)
        self.add_object(s_obj_pngmerge16, png_merger_4bpp)

        png_merger_8bpp = PngMerger_8BPP()
        png_merger_8bpp.set_parent(self)
        self.add_object(s_obj_pngmerge256, png_merger_8bpp)

        png_shape_draw_4bpp = PngShapeDraw_4BPP()
        png_shape_draw_4bpp.set_parent(self)
        self.add_object(s_obj_pngshapedraw16, png_shape_draw_4bpp)

        # * file type descriptions *

        self.add_file_ext_descr(ext_subf_anm, file_ext_descr_anm)
        self.add_file_ext_descr(ext_subf_bmp, file_ext_descr_bmp)
        self.add_file_ext_descr(ext_subf_fnt, file_ext_descr_fnt)
        self.add_file_ext_descr(ext_subf_pal, file_ext_descr_pal)
        self.add_file_ext_descr(ext_subf_scr, file_ext_descr_scr)

        # * decoder objects *

        # FIXME, has no "name", unlike ViewerDataObj
        #   (but the "name" is not an arg, stored in cl.)
        self.add_decoder(ext_subf_pal, SubFileDecoder_Pal_T12())
        self.add_decoder(ext_subf_pal, SubFileDecoder_Pal_T3())
        self.add_decoder(ext_subf_bmp, SubFileDecoder_Bmp_T12())
        self.add_decoder(ext_subf_bmp, SubFileDecoder_Bmp_T3())
        self.add_decoder(ext_subf_bmp, SubFileDecoder_BmpBin())
        self.add_decoder(ext_subf_anm, SubFileDecoder_Anm_T2())
        self.add_decoder(ext_subf_fnt, SubFileDecoder_Fnt_1BPP())
        self.add_decoder(ext_subf_scr, SubFileDecoder_Scr())

        # TODO TODO: add more

        # * ViewerDataObj objects *

        self.add_viewer_data_obj(ext_subf_pal, ViewerDataObj_Pal())
        self.add_viewer_data_obj(ext_subf_bmp, ViewerDataObj_Bmp())
        self.add_viewer_data_obj(ext_subf_fnt, ViewerDataObj_Fnt())
        self.add_viewer_data_obj(ext_subf_anm, ViewerDataObj_AnmFrames())
        self.add_viewer_data_obj(ext_subf_scr, ViewerDataObj_Scr())

        viewer_stored_data = ViewerDataObj_StoredData()
        viewer_stored_data.set_stored_data(info_checksums,
                                           info_frame_sub_image_pos)
        self.add_viewer_data_obj(s_res_file_special, viewer_stored_data)

        # TODO TODO: add more

        # * exporter objects *

        self.add_exporter(SingleBinFileExporter())
        self.add_exporter(AllBinFilesExporter())
        self.add_exporter(BinFileListExporter())
        self.add_exporter(BinFilePatternExporter())
        self.add_exporter(ImageSingleSubItemExporter())
        self.add_exporter(ImageSubItemsExporter())
        self.add_exporter(ImageListSubItemsExporter())
        self.add_exporter(ImageFilePatternSubItemsExporter())
        self.add_exporter(PlainNameListPatternExporter())
        self.add_exporter(PlainSubItemNameListExporter())
        self.add_exporter(FileInfoCsvExporter())
        self.add_exporter(ItemInfoCsvExporter())

        # * resource file open callbacks *
        self.add_open_res_callback_func(s_open_res_cback_pal,
                                        open_res_callback_pal)
