#!/usr/bin/python3
# coding: utf8

# The Incredible Machine (TIM) image resource reader, version 0.1
# By knt47, licensed under the Creative Commons Zero license

from __future__ import division, print_function, unicode_literals

import itertools, sys

if sys.version_info[0] != 3:
    raise NotImplementedError('Only Python 3 is supported.')
else:
    izip = zip
    xrange = range

import zlib
from collections import namedtuple
from timgres.base.misc import (print_msg, print_traceback, namedtuple_to_str,
                               s_vdata_anmframes, s_vdata_anmviewer,
                               s_obj_pngshapedraw16, s_opt_flip,
                               s_res_file_special, ext_subf_anm)
from timgres.base.vdobase import (ViewerDataObj_MultiSubImage,
                                  ViewerDataObj_Animation, nt_info_col_header,
                                  nt_size, flip_none, _num_list_to_str_nosp)

# --- ResDataBase_ViewerDataObj subclass for .ANM files  ---

'''
This module contailns two viewer classes for animations stored in .ANM files:
    ViewerDataObj_AnmFrames: for viewing the animation frames without
    animation and ViewerDataObj_AnmViewer for animation.
'''

# FIXME, these constants should be moved to ...

anm_d_entry_type_sub_img = 2
anm_d_entry_type_rectangle = 3
anm_d_entry_type_line = 4
anm_d_entry_type_sound = 5

# nt_sub_item_info_anmdata: Namedtuple for sub-item information (sub_item_info)
#     used in class ViewerDataObj_AnmFrames. All members are strings.
# item_text: Text displayed in the first column in the "Information" panel
#     (table), also used by get_sub_item_texts
# color_num_or_info: Text of the first column in the "information table".
# r, g, b: Information for each color of the palette
# notes: Additional information.

nt_sub_item_info_anmdata = namedtuple(
    'nt_sub_item_info_anmdata',
    ('item_text', 'x_pos', 'y_pos', 'width', 'height', 'notes'))


def _sect_d_type_2_item_info_to_str(nt_info):
    '''Helper function, for conversion of nt_info (type:
        nt_sect_d_entry_type_2) to a string that will be displayed
        in the "Notes" column. (returns str)'''
    return 'bmp_file_id: %d bmp_sub_img_id: %d img_flip: %d' % (
        nt_info.bmp_file_id, nt_info.bmp_sub_img_id, nt_info.img_flip)


class ViewerDataObj_AnmFrames(ViewerDataObj_MultiSubImage):
    '''
    ViewerDataObj_MultiSubImage subclass, for viewing animation frames,
        that uses the data stored in .ANM animation files.
    '''

    # obj_identifier, obj_description: The ID and the description of the
    #     object (defined in ResDataBase_ViewerDataObj)
    obj_identifier = s_vdata_anmframes
    obj_description = 'ANM animation frames'

    # info_col_headers: corresponds to the items of nt_sub_item_info_anmdata
    info_col_headers = (
        nt_info_col_header('Sub-file', 140),
        nt_info_col_header('X pos.', 50),
        nt_info_col_header('Y pos.', 50),
        nt_info_col_header('Width*', 50),
        nt_info_col_header('Height*', 50),
        nt_info_col_header('Notes', 560),
    )

    def reset_data_subclass(self):
        '''
        Re-implementation of reset_data_subclass from
            ResDataBase_ViewerDataObj. Initializes class-specific fields.
        '''
        # anm_data: nt_anm_data namedtuple (currently, no type check)
        self.anm_data = None

    def get_anm_data(self):
        '''
        Returns the result of the ANM decoder, sotred in the object
            (type: nt_anm_data)
        '''
        return self.anm_data

    def set_sub_image_load_info(self):
        '''
        Re-implementation of set_sub_image_load_info from
            ViewerDataObj_MultiSubImage.
        '''
        # set_sub_image_load_info is always called before set_frame_data, so
        #     the ANM file is decoded here

        # TODO, no type check
        # (NOTE: at present, ANM decode only accepts None a process info)
        self.anm_data = self.parent.decode_sub_file(self.sub_file_name, None)

        # only section D (anm_data_d) and bmp_names is used here
        for sect_d_sub_list in self.anm_data.anm_data_d:
            for sect_d_item in sect_d_sub_list:
                if sect_d_item.entry_type == anm_d_entry_type_sub_img:
                    # Note: one-baseed indices!
                    # img_flip is usable without modification
                    bmp_name = self.anm_data.bmp_names[ \
                        sect_d_item.bmp_file_id - 1].upper()

                    self.add_single_sub_image_load_info(
                        bmp_name, sect_d_item.bmp_sub_img_id,
                        sect_d_item.img_flip)

    def set_frame_data(self):
        '''
        Re-implementation of set_frame_data from ViewerDataObj_MultiSubImage.
        Note that several animations contain instructions for drawing a
            rectangle or a line. These are drawn to PNG images, using
            the PngShapeDraw_4BPP object in parent.
        '''
        # NOTE: this class ignores information in sections A and C!

        if len(self.anm_data.anm_data_b) != len(self.anm_data.anm_data_d):
            raise RuntimeError('Incompatible B and D sections '
                               'detected in self.anm_data')

        if not self.anm_data.anm_data_b: return

        pos_x_min = min(x.pos_x for x in self.anm_data.anm_data_b)
        pos_y_min = min(x.pos_y for x in self.anm_data.anm_data_b)
        # HACK, width is (maybe) not "enough" for some anim. frames
        width_max = (max(x.width
                         for x in self.anm_data.anm_data_b) - pos_x_min)
        # HACK, height is not "enough" for some anim. frames
        height_max = (max(x.height
                          for x in self.anm_data.anm_data_b) - pos_y_min)

        png_shape_draw = self.parent.get_object(s_obj_pngshapedraw16)

        for sect_b_item, sect_d_sub_list in izip(self.anm_data.anm_data_b,
                                                 self.anm_data.anm_data_d):

            cnt_rects = 0
            cnt_lines = 0
            list_sounds = []
            sect_b_pos_x = sect_b_item.pos_x - pos_x_min
            sect_b_pos_y = sect_b_item.pos_y - pos_y_min

            # self.add_new_frame(sect_b_item.width - pos_x_min,
            #                    sect_b_item.height - pos_y_min)

            # self.add_new_frame(sect_b_item.width, sect_b_item.height)
            # HACK for correct positioning when the image is flipped
            self.add_new_frame(width_max, height_max)

            for sect_d_item in sect_d_sub_list:
                if sect_d_item.entry_type == anm_d_entry_type_sub_img:
                    bmp_name = self.anm_data.bmp_names[ \
                        sect_d_item.bmp_file_id - 1].upper()

                    self.add_new_img_curr_frame(
                        bmp_name, sect_d_item.bmp_sub_img_id,
                        sect_d_item.img_flip, sect_d_item.pos_x + sect_b_pos_x,
                        sect_d_item.pos_y + sect_b_pos_y)

                elif sect_d_item.entry_type == anm_d_entry_type_rectangle:
                    # Rectangle is filled when the most significant bit
                    #     of fill is set!

                    png_data_rect = png_shape_draw.draw_rectangle_with_border(
                        sect_d_item.width, sect_d_item.height,
                        sect_d_item.border_color,
                        (sect_d_item.fill & 0xff if sect_d_item.fill
                         & 0x8000 else None))

                    self.add_single_sub_img_data_special(
                        png_data_rect, sect_d_item.pos_x - pos_x_min,
                        sect_d_item.pos_y - pos_y_min)

                    cnt_rects += 1

                elif sect_d_item.entry_type == anm_d_entry_type_line:
                    png_data_line = png_shape_draw.draw_line(
                        sect_d_item.pos_x2 - sect_d_item.pos_x1,
                        sect_d_item.pos_y2 - sect_d_item.pos_y1,
                        sect_d_item.line_color)

                    self.add_single_sub_img_data_special(
                        png_data_line,
                        min(sect_d_item.pos_x1, sect_d_item.pos_x2) -
                        pos_x_min,
                        min(sect_d_item.pos_y1, sect_d_item.pos_y2) -
                        pos_y_min)

                    cnt_lines += 1

                elif sect_d_item.entry_type == anm_d_entry_type_sound:
                    list_sounds.append(sect_d_item.sound_info)
                    print_msg(
                        ('The .ANM file contains an instruction to play '
                         'sound %d. Sound playback is not implemented.') %
                        sect_d_item.sound_info)
                else:
                    raise RuntimeError('Invalid entry type in .ANM section D')

            # NOTE: Item numbering is somewhat different thian in vd_bmp
            #     only the items "present" are numbered, not all [???]

            list_notes = [self.sub_img_usage_data_text()]

            if cnt_rects:
                list_notes.append('rectangles:')
                list_notes.append(str(cnt_rects))

            if cnt_lines:
                list_notes.append('lines:')
                list_notes.append(str(cnt_lines))

            if list_sounds:
                list_notes.append('sound:')
                list_notes.append(_num_list_to_str_nosp(list_sounds))

            self.end_frame_add_info(
                nt_sub_item_info_anmdata(
                    'Animation frame [%d/%d]' %
                    (len(self.img_data_frames), len(self.anm_data.anm_data_b)),
                    str(sect_b_item.pos_x), str(sect_b_item.pos_y),
                    str(sect_b_item.width), str(sect_b_item.height),
                    ' '.join(list_notes)))

    def set_head_info(self):
        '''
        Re-implementation of set_head_info from ViewerDataObj_MultiSubImage.
        '''
        self.item_info_head = nt_sub_item_info_anmdata(
            'Info: ' + self.obj_description, '', '', '', '',
            namedtuple_to_str(self.anm_data.anm_info))

        # TODO, add section A info

        for index_b, sect_b_item in enumerate(self.anm_data.anm_data_b):
            self.item_info_extra.append(
                nt_sub_item_info_anmdata(
                    'B: Anim. frame %d' % index_b, str(sect_b_item.pos_x),
                    str(sect_b_item.pos_y), str(sect_b_item.width),
                    str(sect_b_item.height),
                    'pos_sect_d: %d' % sect_b_item.offs_sect_d))

        # Todo, add short description on sect. C (instead of Item)

        for index_c, sect_c_item in enumerate(self.anm_data.anm_data_c):
            self.item_info_extra.append(
                nt_sub_item_info_anmdata('C: Item %d' % index_c, '', '', '',
                                         '', namedtuple_to_str(sect_c_item)))

        for index_d, sect_d_sub_list in enumerate(self.anm_data.anm_data_d):
            for index_d_sub, sect_d_item in enumerate(sect_d_sub_list):

                if sect_d_item.entry_type == anm_d_entry_type_sub_img:
                    # fixme, pos_x and pos_y shown twice
                    self.item_info_extra.append(
                        nt_sub_item_info_anmdata(
                            'D: Sub-image %d, %d' % (index_d, index_d_sub),
                            str(sect_b_item.pos_x), str(sect_b_item.pos_y), '',
                            '', _sect_d_type_2_item_info_to_str(sect_d_item)))

                elif sect_d_item.entry_type == anm_d_entry_type_rectangle:
                    self.item_info_extra.append(
                        nt_sub_item_info_anmdata(
                            'D: Rectangle %d, %d' % (index_d, index_d_sub),
                            str(sect_d_item.pos_x), str(sect_d_item.pos_y),
                            str(sect_d_item.width), str(sect_d_item.height),
                            'border: %02x fill: %04x' %
                            (sect_d_item.border_color, sect_d_item.fill)))

                elif sect_d_item.entry_type == anm_d_entry_type_line:
                    self.item_info_extra.append(
                        nt_sub_item_info_anmdata(
                            'D: Line (*: x2, y2) %d, %d' %
                            (index_d, index_d_sub), str(sect_d_item.pos_x1),
                            str(sect_d_item.pos_y1), str(sect_d_item.pos_x2),
                            str(sect_d_item.pos_y2),
                            'color: %02x' % (sect_d_item.line_color, )))

                elif sect_d_item.entry_type == anm_d_entry_type_sound:
                    self.item_info_extra.append(
                        nt_sub_item_info_anmdata(
                            'D: Sound %d, %d' % (index_d, index_d_sub), '', '',
                            '', '', 'sound: %d' % sect_d_item.sound_info))


str_msg_frame_numbering = ('ViewerDataObj_AnmViewer - '
                           'Inconsistency in animation frame numbering')


class ViewerDataObj_AnmViewer(ViewerDataObj_Animation):
    '''
    ViewerDataObj_Animation subclass, for viewing animations contained in
        an ANM file. The animation frame information is stored in an
        associated ViewerDataObj_AnmFrames object.
    '''
    # obj_identifier, obj_description: The ID and the description of the
    #     object (defined in ResDataBase_ViewerDataObj)
    obj_identifier = s_vdata_anmviewer
    obj_description = 'ANM animations'

    # viewer_obj_id_frames, viewer_obj_file_ext: Required for accessing
    #     the associated viewer object that contains the frame data.
    #     (ViewerDataObj_StoredData) (originally defined in the parent)
    viewer_obj_id_frames = s_vdata_anmframes
    viewer_obj_file_ext = ext_subf_anm

    def get_expected_frame_number(self):
        '''
        Re-implementation of get_expected_frame_number,
            from ViewerDataObj_Animation
        '''
        anm_data = self.obj_frame_data.get_anm_data()
        if anm_data is None: return 0

        return len(anm_data.anm_data_d)

    def add_anim_info(self):
        '''
        Re-implementation of add_anim_info, from ViewerDataObj_Animation
        '''
        anm_data = self.obj_frame_data.get_anm_data()

        if anm_data is None: return

        index_sect_c_next = 0
        len_sect_c = len(anm_data.anm_data_c)

        if len_sect_c == 0: return

        # NOTE that a dict comprehension cannot be used here
        #     as we need the first rather than the last item for each
        if anm_data.anm_data_a:
            data_a_dict = {}
            for cnt_frame_a, frame_num_a in enumerate(anm_data.anm_data_a):
                if frame_num_a not in data_a_dict:
                    data_a_dict[frame_num_a] = cnt_frame_a
        else:
            data_a_dict = None

        for cnt_frame in xrange(len(anm_data.anm_data_d)):
            cnt_frame_anm = (data_a_dict[cnt_frame]
                             if data_a_dict is not None else cnt_frame)

            if index_sect_c_next < len_sect_c:
                first_frame_sect_c_next = anm_data.anm_data_c[
                    index_sect_c_next].first_frame_id

                if first_frame_sect_c_next == cnt_frame_anm:
                    if index_sect_c_next != 0:
                        self.end_current_animation(
                            anm_data.anm_data_c[index_sect_c_next - 1].anim_id,
                            len_sect_c)

                    index_sect_c_next += 1

                if first_frame_sect_c_next < cnt_frame_anm:
                    if first_frame_sect_c_next == (cnt_frame_anm - 1):
                        # HACK fix for some .ANM files, NOTE: ignores sect. A
                        self.add_anim_frame(cnt_frame - 1, 1)
                        continue

                    print_msg(str_msg_frame_numbering)
                    return

            # NOTE on section A: This section is either empty or an array of 16-bit
            #     integer numbers, in ascending orer from 0 to the number to the
            #     number of animation frames. The only exception from the rule is
            #     that some numbers are repeated (e.g. 0 1 1 1 2...), indicating
            #     which frames should be displayed longer(?)

            # NOTE: cnt_frame, not cnt_frame_anm
            self.add_anim_frame(cnt_frame,
                                (anm_data.anm_data_a.count(cnt_frame)
                                 if anm_data.anm_data_a else 1))

        if index_sect_c_next != len_sect_c:
            # TODO, raises may occur in some "valid" AMN files
            #     So d not raise an exception
            print_msg(str_msg_frame_numbering)
            ## return
            ## raise RuntimeError('Inconsistency in animation frame numbering')

        self.end_current_animation(
            anm_data.anm_data_c[index_sect_c_next - 1].anim_id, len_sect_c)
