#!/usr/bin/python3
# coding: utf8

# The Incredible Machine (TIM) image resource reader, version 0.1
# By knt47, licensed under the Creative Commons Zero license

from __future__ import division, print_function, unicode_literals

import itertools, sys

if sys.version_info[0] != 3:
    raise NotImplementedError('Only Python 3 is supported.')
else:
    izip = zip
    xrange = range

import zlib
from collections import namedtuple
from timgres.base.misc import (print_msg, print_traceback, namedtuple_to_str,
                               s_vdata_special_storeddata,
                               s_vdata_special_storeddata_anim, s_opt_flip,
                               s_res_file_special)
from timgres.base.vdobase import (ViewerDataObj_MultiSubImage,
                                  ViewerDataObj_Animation, nt_info_col_header,
                                  nt_size, flip_none, _num_list_to_str_nosp)

# --- Viewer objs for "stored data" (information not from game resources) ---

'''
This module contains two ViewerDataObj classes, ViewerDataObj_StoredData and
    ViewerDataObj_StoredDataAnim. These classes are used in the cases where
    the "sub-image" information cannot be read directly from the game
    resource files, but some other "source" is required.
'''

# nt_sub_item_info_storeddata: Namedtuple for sub-item information
#     (sub_item_info), used in class ViewerDataObj_StoredData.
#     All members are strings.
# item_text: Text displayed in the first column in the "Information" panel
#     (table), also used by get_sub_item_texts
# width, height: Dimensions of the animation frame.
# notes: Additional information.
nt_sub_item_info_storeddata = namedtuple(
    'nt_sub_item_info_storeddata', ('item_text', 'width', 'height', 'notes'))

str_err_storeddata_nonearg = ('set_stored_data: '
                              'Accepts either 0 or 2 None arguments.')

str_warn_group_unavail = ('ViewerDataObj_StoredData: Warning: %s, '
                          'items in group %d will not be available')

# See documentation in reset_stored_data
index_sub_imgs_in_data = 2
name_pattern_res_part = 'PART%d.BMP'


class ViewerDataObj_StoredData(ViewerDataObj_MultiSubImage):
    '''
    A ViewerDataObj_MultiSubImage subclass that supports animation frames
        composed of multiple images, but no animation. This class does
        not read the animation frame information form a game resource
        file, but depends on a "data source" (see set_stored_data)
    '''

    # obj_identifier, obj_description: The ID and the description of the
    #     object (defined in ResDataBase_ViewerDataObj)
    obj_identifier = s_vdata_special_storeddata
    obj_description = 'Animation frames (TIM/TEMIM), multi-part'

    # info_col_headers: corresponds to the items of nt_sub_item_info_storeddata
    info_col_headers = (
        nt_info_col_header('Sub-file', 160),
        nt_info_col_header('Width', 60),
        nt_info_col_header('Height', 60),
        nt_info_col_header('Notes', 620),
    )

    def __init__(self):
        '''Constructor of the class. Requires no argument.'''
        ViewerDataObj_MultiSubImage.__init__(self)
        self.reset_stored_data()

    def reset_stored_data(self):
        '''
        Resets the "stored data", information about the animation frames.
        '''
        # NOT affected by reset_data. Format description copied from t1_p1info.

        # data_checksums: The check-sums used for identifying the game version
        # A tuple, one item for each group (depending on presence/absence in)
        #     versions
        # Format if "inner" tuples:
        #     0. group id (based on availability in demo/full game), int
        #     1. image resource name (str)
        #     2. the crc32 checksum (unsigned 32 bit int)
        self.data_checksums = None

        # data_sub_img_pos: The positions (x, y) & other important data
        #     for each animation frame
        # Format of inner tuples:
        #     0. Frame information tuple, see fromat below
        #     1... information for each sub-image that will be displayed
        #
        # Format of "inner-inner" tuple 0 (frame information)
        #     0. The group id (from info_checksums)
        #     1. animation information
        #     2. frame width
        #     3. frame height
        #     4. x offset (frame)
        #     5. y offset (frame)
        #
        # Format of "inner-inner" tuples 1... (info for each sub-image):
        #     0. Resource file I.D eg. 15 -> 'PART15.BMP'
        #     1. Resource sub-item ID
        #     2. x offset (sub-image)
        #     3. y offset (sub-image)
        #     4. reserved for flip, currently unused
        self.data_sub_img_pos = None

    def reset_data_subclass(self):
        '''
        Re-implementation of reset_data_subclass from
            ResDataBase_ViewerDataObj. Initializes class-specific fields.
        '''
        # groups_present: Available Group IDs, the crc32 check was successful.
        self.groups_present = set()

        # sub_item_count: set in set_stored_data, required by set_frame_data
        #     before self.img_data_frames is "complete"
        self.sub_item_count = 0

    def set_stored_data(self, new_data_checksums, new_data_sub_img_pos):
        '''
        Sets the "stored data" for the class (sub-image and checksum
            information).
        new_data_checksums: List or tuple, see description in vd_sdata.py
        new_data_sub_img_pos: List or tuple, see description in vd_sdata.py
        '''
        if new_data_checksums is not None:
            if not isinstance(new_data_checksums, (tuple, list)):
                raise TypeError('new_data_checksums: must be list or tuple')

            if new_data_sub_img_pos is None:
                raise ValueError(str_err_storeddata_nonearg)

        if new_data_sub_img_pos is not None:
            if not isinstance(new_data_sub_img_pos, (tuple, list)):
                raise TypeError('new_data_sub_img_pos: must be list or tuple')

            if new_data_checksums is None:
                raise ValueError(str_err_storeddata_nonearg)

        self.data_checksums = new_data_checksums
        self.data_sub_img_pos = new_data_sub_img_pos

    def get_data_sub_img_pos(self):
        '''
        For internal use only, returns a copy of self.data_sub_img_pos
            (list or tuple, format described in vd_sdata.py)
        '''
        return self.data_sub_img_pos

    def get_groups_present(self):
        '''
        For internal use only, returns a copy of self.groups_present,
            the available groups, based on checksum check (type: set).
        '''
        return set(self.groups_present)

    def get_item_count_storeddata(self):
        '''
        For internal use only, returns a copy of self.sub_item_count
            (type: int)
        '''
        return self.sub_item_count

    def set_sub_image_load_info(self):
        '''
        Re-implementation of set_sub_image_load_info from
            ViewerDataObj_MultiSubImage. Performs two steps: 1.
            Performs the crc32 checks, using the data in self.data_checksums
            2. Calls add_single_sub_image_load_info for each required
            sub-images.
        '''

        # As set_sub_image_load_info is always called before set_frame_data,
        #     crc32 checks are done in this method

        #  do nothing if data_ fields are None
        if (self.data_checksums is None) or (self.data_sub_img_pos is None):
            return

        # ...  crc32 checks ...
        res_read = self.parent.get_res_read_obj()

        for group_id, sub_f_name, crc32_correct in self.data_checksums:
            try:
                file_data = res_read.get_single_file_data(sub_f_name)
            except ValueError as ex:
                # get_single_file_data raises ValueError if
                #     the sub-file does not exist
                print_traceback()
                warn_msg_part = 'Sub-file "%s" not found' % sub_f_name
                print_msg(str_warn_group_unavail % (warn_msg_part, group_id))
                # No print_traceback!
                continue

            # zlib.crc32() & 0xffffffff: Python 2 compat hack
            if (zlib.crc32(file_data) & 0xffffffff) == crc32_correct:
                self.groups_present.add(group_id)
            else:
                warn_msg_part = 'Wrong crc32 checksum, file "%s"' % sub_f_name
                print_msg(str_warn_group_unavail % (warn_msg_part, group_id))

        # fixme, duplication, set_frame_data
        for sub_item_data in self.data_sub_img_pos:
            # sub_item_data[0] is not unpacked here, only one item needed
            if sub_item_data[0][0] not in self.groups_present:
                continue

            for res_f_num, sub_img_id, offs_x, offs_y, flip in \
                    sub_item_data[1:]:
                # Note_ offs_x, offs_y unused here
                self.add_single_sub_image_load_info(
                    name_pattern_res_part % res_f_num, sub_img_id, flip)

            self.sub_item_count += 1

    def set_frame_data(self):
        '''
        Re-implementation of set_frame_data from ViewerDataObj_MultiSubImage.
        '''

        for sub_item_data in self.data_sub_img_pos:
            # sub_item_data[0][0] is the group ID

            group_num, anim_info, frame_w, frame_h, frame_offs_x, \
            frame_offs_y = sub_item_data[0]

            if group_num not in self.groups_present: continue

            self.add_new_frame(frame_w, frame_h)

            for res_f_num, sub_img_id, offs_x, offs_y, flip in sub_item_data[
                    1:]:
                self.add_new_img_curr_frame(name_pattern_res_part % res_f_num,
                                            sub_img_id, flip,
                                            offs_x + frame_offs_x,
                                            offs_y + frame_offs_y)

            # NOTE: Item numbering is somewhat different thian in vd_bmp
            #     only the items "present" are nu,bered, not the full
            #      self.data_sub_img_pos
            self.end_frame_add_info(
                nt_sub_item_info_storeddata(
                    'Animation frame (TIM/TEMIM) [%d/%d]' %
                    (len(self.img_data_frames), self.sub_item_count),
                    str(frame_w), str(frame_h),
                    self.sub_img_usage_data_text()))

    def set_head_info(self):
        '''
        Re-implementation of set_head_info from ViewerDataObj_MultiSubImage.
        '''
        self.item_info_head = nt_sub_item_info_storeddata(
            'Info: ' + self.obj_description, '', '',
            ('groups: %s' % _num_list_to_str_nosp(self.groups_present)))


# TODO, copied from t1p1nfo (a module that intentionally does not
#     import it from some other source file)
anim_unused = 0
anim_first_frame = 1
anim_continue = 2
anim_first_flipped = 3


class ViewerDataObj_StoredDataAnim(ViewerDataObj_Animation):
    '''
    A ViewerDataObj_Animation subclass for viewing animations.
    The data required for the animation is stored in an 'associated"
        ViewerDataObj_StoredData.
    '''
    # obj_identifier, obj_description: the identifier and a short description
    #     of the viewer class (ResDataBase_ViewerDataObj)
    obj_identifier = s_vdata_special_storeddata_anim
    obj_description = 'Animation (TIM/TEMIM)'

    # viewer_obj_id_frames, viewer_obj_file_ext: Required for accessing
    #     the associated viewer object that contains the frame data.
    #     (ViewerDataObj_StoredData) (originally defined in the parent)
    viewer_obj_id_frames = s_vdata_special_storeddata
    viewer_obj_file_ext = s_res_file_special

    def get_expected_frame_number(self):
        '''
        Re-implementation of get_expected_frame_number,
            from ViewerDataObj_Animation.
        '''
        return self.obj_frame_data.get_item_count_storeddata()

    def add_anim_info(self):
        '''
        Re-implementation of add_anim_info, from ViewerDataObj_Animation.
        Uses the data from the associated viewer object
            (s_vdata_special_storeddata)
        '''
        data_sub_img_pos = self.obj_frame_data.get_data_sub_img_pos()
        if data_sub_img_pos is None: return
        groups_present = self.obj_frame_data.get_groups_present()
        item_count = self.obj_frame_data.get_item_count_storeddata()
        cnt_anim = -1

        for sub_item_id, sub_item_data in enumerate(data_sub_img_pos):
            group_num = sub_item_data[0][0]
            if group_num not in groups_present: continue

            anim_info = sub_item_data[0][1]
            if anim_info == anim_unused: continue

            # TODO: currently anim_first_frame and anim_first_flipped
            #     behave the same way

            if anim_info in (anim_first_frame, anim_first_flipped):
                if cnt_anim != -1:
                    self.end_current_animation(cnt_anim, item_count)

                cnt_anim += 1

            if cnt_anim == -1:
                raise RuntimeError('Invalid "stored" data, first item must '
                                   'use anim_first_frame, not anim_continue')

            # Duration other than 1 is currently not supported
            self.add_anim_frame(sub_item_id, 1)

        self.end_current_animation(cnt_anim, item_count)
